#include ".\camera.h"
const float Camera::M_PI=3.14159265f;

Camera::Camera(void) { }

Camera::~Camera(void) { }

void Camera :: setModelViewMatrix(void) {
	// load modelview matrix with existing camera values
	float m[16];
	Vector3 eVec(eye.x, eye.y, eye.z); // a vector version of eye 
	m[0] = u.x;		m[4] = u.y;		m[8]  = u.z;	m[12] = -eVec.dot(u);
	m[1] = v.x;		m[5] = v.y;		m[9]  = v.z;	m[13] = -eVec.dot(v);
	m[2] = n.x;		m[6] = n.y;		m[10] = n.z;	m[14] = -eVec.dot(n);
	m[3] = 0;		m[7] = 0;		m[11] = 0;		m[15] = 1.0;
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	glLoadMatrixf(m); // load OpenGLs modelview matrix
}

void Camera:: set(Point3 Eye, Point3 look, Vector3 up) {
	// create a modelview matrix and send it to OpenGL
	eye.set(Eye); // store the given eye position
	n.set(eye.x - look.x, eye.y - look.y, eye.z - look.z); // make n
	u.set(up.cross(n)); // make u = up X n
	n.normalize(); u.normalize(); // make them unit length
	v.set(n.cross(u));  // make v =  n X u
	setModelViewMatrix(); // tell OpenGL 
}

//void Camera:: setOblique(Vector3 d,..others..) { // establish camera for oblique projections
//	glMatrixMode(GL_PROJECTION);
//	glLoadIdentity();
//	glOrtho(l,r,b,t,n,f);         // set the projection matrix
//	if(d.z == 0.0) return;		// for orthographic projections
//	float m[16]; // space for a 4 by 4 matrix
//	for(int i = 0; i < 16; i++)   // start with identity matrix
//	    m[i] = (i%5 == 0)? 1.0 : 0.0;// identity matrix
//	m[8] = -d.x/d.z;              // add the shear terms
//	m[9] = -d.y/d.z;
//	glMultMatrixf(m);             // postmultiply it by m
//}

void Camera :: roll(float angle) { // roll the camera through angle degrees
	float cs = cos(3.14159265f/180.0f * angle);
	float sn = sin(3.14159265f/180.0f * angle);
	Vector3 t(u); // remember old u
	u.set(cs*t.x - sn*v.x, cs*t.y - sn*v.y, cs*t.z - sn*v.z);
	v.set(sn*t.x + cs*v.x, sn*t.y + cs*v.y, sn*t.z + cs*v.z);
	setModelViewMatrix();
}

void Camera:: raytrace(Scene &scn, int blockSize) {		// setShape MUST be called prior to this call
	Ray theRay;
	Color3 clr;
	theRay.setStart(eye);
	// set up OpenGL for simple 2D drawing
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	int nCols = 640;
	int nRows = 480;
	gluOrtho2D(0,nCols,nRows,0);	// whole screen is the window		I had the lsat 2 vars switched
	glDisable(GL_LIGHTING);	// so glColor3f works properly
	// begin raytracing
	int numRays=0;
	for (int row=0; row<nRows; row += blockSize) {
		float y = -H + row * 2 * H / (float)nRows;
		for (int col=0; col<nCols; col += blockSize) {
			float x = -W + col * 2 * W / (float)nCols;
			theRay.setDir(-(float)nearDist*n.x + x*u.x + y*v.x, -(float)nearDist*n.y + x*u.y + y*v.y,
						  -(float)nearDist*n.z + x*u.z + y*v.z);	// set the ray's direction
			//theRay.dir.normalize();
			//clr.set(scn.shade(theRay));	// find the color
			clr = scn.shade(theRay);	// find the color
			glColor3f(clr.red, clr.green, clr.blue);
			glRecti(col, row, col+blockSize, row+blockSize);
			numRays++;
		}
	}
	if (numRays)
		numRays=0;
}

void Camera::slide(float delU, float delV, float delN) {
	eye.x += delU*u.x + delV*v.x + delN*n.x;
	eye.y += delU*u.y + delV*v.y + delN*n.y;
	eye.z += delU*u.z + delV*v.z + delN*n.z;
	setModelViewMatrix();
}
void Camera::setShape(float vAng, float asp, float nearD, float farD) {
	viewAngle = vAng;
	aspect = asp;
	nearDist = nearD;
	farDist = farD;
	H = (float)(nearDist*tan(viewAngle*M_PI/360));//nCols/2;
	W = H*(float)aspect;//nRows/2;
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	gluPerspective(viewAngle, aspect, nearDist, farDist);
}