#pragma once
#include <string>
#include <iostream>
#include <fstream>
#include <strstream>
#include "gxbase.h"
using namespace std;
#include "RGBpixmap.h"
#include "starters.h"
class GeomObj;

// Shapes class and Supporting classes
// Shapes.h 
// GeomObj, Boolean, Shape, Cube, Sphere, etc.
// uses pixmap class and basic classes above

#define PI 3.14159265
#define TWOPI (3.14159265 * 2)
#define degPerRad (180/3.14159265)
#define Near(a,v) fabs((a)-(v))<0.0001

//@@@@@@@@@@@@@@@@@@@ PointCluster class @@@@@@@@@@@@@@@@@@
class PointCluster { // holds array of points for the bounding hull of a shape
public:
	int num;
	Point3* pt;
	PointCluster() { num = 0; pt = NULL; }
	PointCluster(int n) { // make a cluster of n points
		pt = new Point3[n]; assert(pt);
		num = n;
	}
};

//@@@@@@@@@@@@@@@@@@@@@@@@@@ SphereInfo @@@@@@@@@@@@@@@@@@@@@
class SphereInfo { // holds the center and radius of a sphere
public:
	Point3 center;
	float radSq;
	void set(float x, float y, float z, float rsq) {
		center.set(x,y,z);
		radSq = rsq;
	}
};

//@@@@@@@@@@@@@@@@@@@@@@@@ Cuboid @@@@@@@@@@@@@@@@@@@@@@@@
class Cuboid { // holds six border values of a cuboid
public:
	float left, top, right, bott, front, back;
	void set(float l, float t, float r, float b, float f, float bk) {
		left = l; top = t; right = r; bott = b; front = f; back = bk;
	}
	void set(Cuboid& c) {
		left = c.left; top = c.top; right = c.right; bott = c.bott;
		front = c.front; back = c.back;
	}
	void print() {
		cout << "\n(l,t,r,b,f,bk) = (" << left << "," << top << "," << right 
			 << "," << bott << "," << front << "," << back << ")";
	}
};

//@@@@@@@@@@@@@@@@@@@@ Ray @@@@@@@@@@@@@@@@@@@@@@@@@
class Ray {
public:
	Point3 start;
	Vector3 dir;
	int recurseLevel;
	int row, col; // for screen extents
	int numInside; // number of objects on list
	GeomObj* inside[10]; // array of object pointers

	Ray() {	// default constructor
		start.set(0,0,0);
		dir.set(0,0,0);
		numInside = 0;
		recurseLevel = 0;
	}
	Ray(Point3 origin) { // constructor: set start point of ray
		start.set(origin);
		dir.set(0,0,0);
		numInside = 0;
		recurseLevel = 0;
	}
	Ray(Point3& origin, Vector3& direction) { // constructor: set start and dir of ray
		start.set(origin); 
		dir.set(direction);
		numInside = 0;
		recurseLevel = 0;
	}
	Point3 rayPos(double time) {
		return Point3(start.x+(float)time*dir.x, start.y+(float)time*dir.y, start.z+(float)time*dir.z);
	}
	Ray xfrmRay(Affine4 invT) {
		return Ray(Point3(start.x*invT.m[0]+start.y*invT.m[1]+start.z*invT.m[2]+invT.m[3],
						start.x*invT.m[4]+start.y*invT.m[5]+start.z*invT.m[6]+invT.m[7],
						start.x*invT.m[8]+start.y*invT.m[9]+start.z*invT.m[10]+invT.m[11]),
						Vector3(dir.x*invT.m[0]+dir.y*invT.m[1]+dir.z*invT.m[2],
						dir.x*invT.m[4]+dir.y*invT.m[5]+dir.z*invT.m[6],
						dir.x*invT.m[8]+dir.y*invT.m[9]+dir.z*invT.m[10]));
	}
	void setStart(Point3& p) { start.set(p); }
	void setDir(float x, float y, float z) { dir.x = x; dir.y = y; dir.z = z; }
	void setRayDirection(Light *L); // for shadow feelers
	void setRayDirection(Vector3& dir); // for spawned rays
	int isInShadow();
	void makeGenericRay(GeomObj* p, Ray& gr);
};

//@@@@@@@@@@@@@@@@@@ HitInfo @@@@@@@@@@@@@@@@@@@@
class HitInfo { // data for each hit with a surface
  public:
	double hitTime;    		// the hit time
	GeomObj* hitObject;  	// the object hit
	int surface;			    // which surface is hit?
	int isEntering;			 // is ray entering the object?
	Point3 hitPoint;		   // hit point 
	Vector3 hitNormal;		// normal at hit point
	HitInfo() 	{ 
		hitObject = NULL;
		hitTime = -1000;
		surface = 0;
		isEntering = 0;
	}
	void set(HitInfo& h) {
		hitTime = h.hitTime; hitObject = h.hitObject;	
		surface = h.surface; isEntering = h.isEntering;
		hitPoint.set(h.hitPoint); hitNormal.set(h.hitNormal);
	}
};

//@@@@@@@@@@@@@@@@@ Intersection @@@@@@@@@@@@@@@@@@@@@
class Intersection { // hold the hit list 
 public:
	#define maxNumHits 8
	int numHits;			 // the number of hits
	HitInfo hit[maxNumHits];	 // list of hits;
	Intersection() { numHits = 0; }  // default constructor
	void set(Intersection& intr) { // copy intersection info
		numHits = intr.numHits;
		for(int i = 0; i < maxNumHits; i++)
			hit[i].set(intr.hit[i]);
	}
};

//@@@@@@@@@@@@@@@@@ Material class @@@@@@@@@@@@@@
class Material {
public:
	Color3 ambient, diffuse, specular, emissive;
	int numParams; // for textures
	float params[10]; // for textures
	int textureType;  // 0 for none, neg for solids, pos for images
	float specularExponent, reflectivity, transparency, speedOfLight;
	float specularFraction, surfaceRoughness;	
	void setDefault() {
		textureType = 0; // for none
		numParams = 0;
		reflectivity = transparency = 0.0;
		speedOfLight = specularExponent = 1.0;
		specularFraction = 0.0;
		surfaceRoughness = 1.0;
		ambient.set(0.1f,0.1f,0.1f);
		diffuse.set(0.8f,0.8f,0.8f);
		specular.set(0,0,0);
		emissive.set(0,0,0);
	}
	void set(Material& m) {
		textureType = m.textureType;
		numParams = m.numParams;
		for(int i = 0; i < numParams; i++) params[i] = m.params[i];
		transparency = m.transparency;
		speedOfLight = m.speedOfLight;
		reflectivity = m.reflectivity;
		specularExponent = m.specularExponent;
		specularFraction = m.specularFraction;
		surfaceRoughness = m.surfaceRoughness;
		ambient.set(m.ambient);
		diffuse.set(m.diffuse);
		specular.set(m.specular);
		emissive.set(m.emissive);
	}
}; // end of Material

//@@@@@@@@@@@@@@@@@@@@@ GeomObj class @@@@@@@@@@@@@@@@
class GeomObj {
public: 
	IntRect scrnExtnt;
	Cuboid genBoxExtent,worldBoxExtent;
	SphereInfo genSphereExtent,worldSphereExtent;
	GeomObj * next;

	GeomObj(): next(NULL) {}
	virtual ~GeomObj() { if(next) delete(next); next=NULL; }
	virtual bool hit(Ray &r, Intersection &inter) { return false; }
	virtual void loadStuff() {}
	virtual void drawOpenGL() {}
	virtual void tellMaterialsGL() {}
	virtual void makeExtentPoints(PointCluster& clust) { cout << "in makeExtentPoints for GeomObj"; }
	virtual Point2 texturePoint(Point3 p) { Point2 pt(0,0); return pt; }
	virtual Color3 texture(HitInfo& h) { Color3 c(0,0,0); return c; }
};

//@@@@@@@@@@@@@@@@@@@ Boolean @@@@@@@@@@@@@@@@@@
class Boolean: public GeomObj {
public:
	GeomObj *left, *right;
	Boolean():left(NULL),right(NULL){}
	virtual bool hit(Ray &r, Intersection &inter){return false;}
	virtual void makeExtentPoints(PointCluster& clust) { cout << "in makeExtentPoints for Boolean"; }
	virtual void drawOpenGL() { // just draw its children
		if(left)left->drawOpenGL();
		if(right)right->drawOpenGL();
	}
};

//@@@@@@@@@@@@@@@@@@@@ UnionBool @@@@@@@@@@@@@@@@
class UnionBool : public Boolean {
public:
	UnionBool() { Boolean(); } //constructor
	virtual void makeExtentPoints(PointCluster& clust) { // cluster is union of lft and rt clusters
		PointCluster lftClust, rtClust;
		left->makeExtentPoints(lftClust);
		right->makeExtentPoints(rtClust);
		// now combine them
		int i,j;
		clust.num = lftClust.num + rtClust.num;
		clust.pt = new Point3[clust.num];
		assert(clust.pt);
		for(i = 0; i < lftClust.num; i++)
			clust.pt[i] = lftClust.pt[i];
		for(j = 0; j < rtClust.num; j++)
			clust.pt[j + i] = rtClust.pt[j];
	}
	virtual bool hit(Ray &r, Intersection &inter);
};

//@@@@@@@@@@@@@@@@@@@@ IntersectionBool @@@@@@@@@@@@@@@@
class IntersectionBool : public Boolean {
public:
	IntersectionBool() { Boolean(); }
	virtual void makeExtentPoints(PointCluster& clust) { // cluster is just left cluster	
		left->makeExtentPoints(clust);
	}
	virtual bool hit(Ray &r, Intersection &inter){return false;}//;
};

//@@@@@@@@@@@@@@@@@@@@ DifferenceBool @@@@@@@@@@@@@@@@
class DifferenceBool : public Boolean {
public:
	DifferenceBool() { Boolean(); } 
	virtual void makeExtentPoints(PointCluster& clust) { // cluster is just left cluster
		left->makeExtentPoints(clust);
	}
	virtual bool hit(Ray &r, Intersection &inter);
};

//@@@@@@@@@@@@@@@@@ Shape @@@@@@@@@@@@@@@@@@@@@
class Shape: public GeomObj {
private:
	bool On(float in[4]) {
		if (in[0] == 0 && in[1] == 0 && in[2] == 0) { return false; }
		return true;
	}
public:
	Material mtrl;
	Affine4 transf, invTransf;
	virtual bool hit(Ray &r, Intersection &inter) { return false; }
	//virtual Color3 texture(HitInfo& h, int whichTexture);
	Shape() { mtrl.textureType = 0; mtrl.numParams = 0; }
	void setMaterial(Material& mt) { mtrl.set(mt); }
	void tellMaterialsGL() {
		float amb[4], diff[4], spec[4], emiss[4];
		float zero[] = {0,0,0,1};
		mtrl.ambient.build4tuple(amb); // fill the array
		mtrl.diffuse.build4tuple(diff);
		mtrl.specular.build4tuple(spec);
		mtrl.emissive.build4tuple(emiss);
		glMaterialfv(GL_FRONT/*_AND_BACK*/,GL_AMBIENT,amb);
		glMaterialfv(GL_FRONT/*_AND_BACK*/,GL_DIFFUSE,diff);
		if(On(spec))
			glMaterialfv(GL_FRONT/*_AND_BACK*/,GL_SPECULAR,spec);
		else 
			glMaterialfv(GL_FRONT,GL_SPECULAR,zero);
		glMaterialfv(GL_FRONT/*_AND_BACK*/,GL_EMISSION,emiss);
		glMaterialf (GL_FRONT/*_AND_BACK*/,GL_SHININESS,mtrl.specularExponent);
	}
	virtual void drawOpenGL() {}
	virtual Point2 texturePoint(Point3 p) { Point2 pt(0,0); return pt; }
	virtual void makeExtentPoints(PointCluster& clust) {}		
}; //end: Shape class

//@$@$@$@$@$@$@$@$@ Cube class $@$@$@$@$@$@$@$@$@$@
class Cube : public Shape {
public:	
	Cube() {}
	void drawOpenGL() {
		tellMaterialsGL();
		glPushMatrix();
			glMultMatrixf(transf.m); //load affine
			glEnable(GL_NORMALIZE);
			glBegin(GL_QUAD_STRIP);
				glVertex3f(-1,-1,1);
				glVertex3f(-1,1,1);
				glVertex3f(1,-1,1);
				glVertex3f(1,1,1);
				glVertex3f(1,-1,-1);
				glVertex3f(1,1,-1);
				glVertex3f(-1,-1,-1);
				glVertex3f(-1,1,-1);
				glVertex3f(-1,-1,1);
				glVertex3f(-1,1,1);
			glEnd();
			glBegin(GL_QUADS);
				glVertex3f(1,1,1);
				glVertex3f(-1,1,1);
				glVertex3f(-1,1,-1);
				glVertex3f(1,1,-1);
			glEnd();
			glBegin(GL_QUADS);
				glVertex3f(1,-1,1);
				glVertex3f(-1,-1,1);
				glVertex3f(-1,-1,-1);
				glVertex3f(1,-1,-1);
			glEnd();
			/*if(doEdges) glutWireCube(2.0);*/
			/*if(doFaces) glutSolidCube(2.0);	// a cube with vertices -1 to +1*/
		glPopMatrix();
	}
	bool hit(Ray &r, Intersection &inter);
	//Point2 texturePoint(Point3 p); // ditto
	void makeExtentPoints(PointCluster& clust){}//; // ditto
	Vector3 cubeNormal(int which) {
		int m = which/2, n = (which%2)?-1:1;
		if (m==0) return Vector3(0,(float)n,0);
		else if (m==1) return Vector3((float)n,0,0);
		return Vector3(0,0,(float)n);
	}
};

//@$@$@$@$@$@$@$@$ Sphere class @$@$@$@$@$@$@$@$@$@$@$@
class Sphere : public Shape {
public:
	void drawOpenGL() {
		GLUquadric * qobj;
		qobj = gluNewQuadric();
		tellMaterialsGL(); 
		glPushMatrix();	
			glMultMatrixf(transf.m); 
			gluSphere(qobj,1.0,20,20);	
		glPopMatrix();
	}
	Sphere() {}
	bool hit(Ray& r, Intersection& inter);	// in SDL.cpp
	//Point2 texturePoint(Point3 p);
	void makeExtentPoints(PointCluster& clust){}//;
};
//@$@$@$@$@$@$@$@$ Circle class @$@$@$@$@$@$@$@$@$@$@$@
class Circle : public Shape {
public:
	void drawOpenGL() {
		GLUquadric * qobj;
		qobj = gluNewQuadric();
		tellMaterialsGL(); 
		glPushMatrix();	
			glMultMatrixf(transf.m); 
			gluDisk(qobj,0.0,1.0,20,20);	
		glPopMatrix();
	}
	Circle() {}
	bool hit(Ray& r, Intersection& inter);	// in SDL.cpp
	//Point2 texturePoint(Point3 p);
	void makeExtentPoints(PointCluster& clust){}//;
};
//@$@$@$@$@$@$@$@$@$@ TaperedCylinder class @$@$@$@$@$@$@$@$@$
class TaperedCylinder : public Shape {
public:
	float smallRadius;
	TaperedCylinder() {}
	bool hit(Ray &r, Intersection &inter);		
	void drawOpenGL() {
		GLUquadric * qobj;
		qobj = gluNewQuadric();
		tellMaterialsGL();
		glPushMatrix();
			glMultMatrixf(transf.m);
			gluCylinder(qobj,1.0,smallRadius,2.0,20,20);	
		glPopMatrix();
	}
	//Point2 texturePoint(Point3 p);
	void makeExtentPoints(PointCluster& clust){}//;
};

//@$@$@$@$@$@$@$@$@$@ Square class @$@$@$@$@$@$@$@$@$
class Square : public Shape {
public:
	Square() {}
	bool hit(Ray &r, Intersection &inter);
	void drawOpenGL() {
		glBegin(GL_QUADS);
			glMultMatrixf(transf.m); 
			glVertex3f(-1,0,1);
			glVertex3f(1,0,1);
			glVertex3f(1,0,-1);
			glVertex3f(-1,0,-1);
		glEnd();
	}
	//Point2 texturePoint(Point3 p);
	void makeExtentPoints(PointCluster& clust){}//;
};

//@$@$@$@$@$@$@$@$@$@ Traingle class @$@$@$@$@$@$@$@$@$
class Triangle : public Shape {
public:
	Triangle() {}
	bool hit(Ray &r, Intersection &inter);
	void drawOpenGL() {
		glBegin(GL_TRIANGLES);
			glMultMatrixf(transf.m); 
			glVertex3f(-1,0,1);
			glVertex3f(1,0,0);
			glVertex3f(-1,0,-1);
		glEnd();
	}
	//Point2 texturePoint(Point3 p);
	void makeExtentPoints(PointCluster& clust){}//;
};

//@$@$@$@$@$@$@$@$@$@ Plane class @$@$@$@$@$@$@$@$@$
class Plane : public Shape {
public:
	Plane() {}
	bool hit(Ray &r, Intersection &inter);
	void drawOpenGL() {
		glBegin(GL_QUADS);
			glMultMatrixf(transf.m); 
			glVertex3f(-1,0,1);
			glVertex3f(1,0,1);
			glVertex3f(1,0,-1);
			glVertex3f(-1,0,-1);
		glEnd();
	}
	//Point2 texturePoint(Point3 p);
	void makeExtentPoints(PointCluster& clust){}//;
};

//################## class VertexID ################  //used to define a Mesh
class VertexID { public: int vertIndex, normIndex; };
//################# class FACE ##############  //used to define a Mesh
class Face {
public:
	int nVerts;
	VertexID * vert; // array of vertex and normal indices
	Face(){ nVerts = 0; vert = NULL;}
	~Face(){delete[] vert; nVerts = 0;}
};
//@$@$@$@$@$@$@$@$@$@ Mesh class @$@$@$@$@$@$@$@$@$
class Mesh : public Shape {
private:
	int numVerts, numNorms, numFaces;
	Point3 *pt; // array of points
	Vector3 *norm; // array of normals
	Face *face; // array of faces
	int lastVertUsed;
	int lastNormUsed;
	int lastFaceUsed;
public:
	void readMesh(string fname);
	void writeMesh(char* fname);
	void printMesh();
	void drawMesh();
	void drawEdges(){}//;
	void drawFrontFaces();
	void freeMesh() { // free up memory used by this mesh.
		delete [] pt; // release whole vertex list
		delete [] norm;
		for(int f = 0; f < numFaces; f++)
			delete[] face[f].vert; // delete the vert[] array of this face
		delete [] face;
	}
	int isEmpty() { return (numVerts == 0) || (numFaces == 0) || (numNorms == 0); }
	void makeEmpty() { numVerts = numFaces = numNorms = 0; }
	Mesh() {
		numVerts = numFaces = numNorms = 0; 
		pt = NULL; norm  =  NULL; face = NULL;
		lastVertUsed = lastNormUsed = lastFaceUsed = -1;
	}
	virtual void drawOpenGL() { 
		tellMaterialsGL();
		glPushMatrix();
		glMultMatrixf(transf.m); 
		/*if (doFaces)*/ drawMesh();	
		/*if (doEdges)*/ drawEdges();
		glPopMatrix();
	} 
	Mesh(string fname){ // read this file to build mesh
		numVerts = numFaces = numNorms = 0; 
		pt = NULL; norm  =  NULL; face = NULL;
		lastVertUsed = lastNormUsed = lastFaceUsed = -1;
		readMesh(fname);
	}
	Vector3 newell4(int indx[]) {
		/* return the normalized normal to face with vertices
		pt[indx[0]],...,pt[indx[3]]. i.e. indx[] contains the four indices 
		into the vertex list to be used in the Newell calculation */
		Vector3 m;
		for(int i = 0; i < 4 ; i++) {
			int next = (i== 3) ? 0 : i + 1; // which index is next?
			int f = indx[i], n = indx[next]; // names for the indices in the pair
			m.x += (pt[f].y - pt[n].y) * (pt[f].z + pt[n].z);
			m.y += (pt[f].z - pt[n].z) * (pt[f].x + pt[n].x);
			m.z += (pt[f].x - pt[n].x) * (pt[f].y + pt[n].y);
		}
		m.normalize();
		return m;
	}
public:
	string meshFileName; // holds file name for this Mesh
	//void print() { cout << "\nMesh "; mtrl.print(); }
	bool hit(Ray &r, Intersection &inter) { return false; }//;
	//Point2 texturePoint(Point3 p);
	void makeExtentPoints(PointCluster& clust){}//;
};
//@@@@@@@@@@@@@@@@@ DefUnit & DefUnitStack classes @@@@@@@@@@@@@@  //used in Scene to read SDL files
class DefUnit {
// developed by Steve Morin
public:
	string    name, stuff;
	DefUnit(string n, string s) {stuff = s;name = n;}
};
class DefUnitStack {
public:
	DefUnitStack() {
		stack = NULL;
	}
	void push(string n, string s) { 
		D4S *temp_d4s = new D4S;
		temp_d4s->current = new DefUnit(n, s);
		temp_d4s->next = stack;
		stack = temp_d4s;
	}
	void print() {
		D4S *temp = stack;
		string t;
		while (temp) {
			cout << temp->current->name << ":" ;
			cout << temp->current->stuff << endl;
			temp = temp->next;
		}
	}
	int search(string s) {
		D4S *temp = stack;
		while (temp) {
			if ( temp->current->name == s ) {
				return(1);
			}
			temp = temp->next;
		}
		return(0);
	}
	string contents(string s) {
		D4S *temp = stack;
		while (temp) {
			if (temp->current->name == s ) {
				return(temp->current->stuff);
			}
			temp = temp->next;
		}
		return(NULL);
	}
	void release() {
		while(stack) {
			D4S* tmp = stack; // grab it
			//cerr << "releasing def_stack item: "<< tmp->current->name<< endl;
			stack = stack->next; // advance p
			delete tmp->current; // release 2 strings
			delete tmp; // release node
		}
		stack = NULL;
	}
private:
	struct D4S {
		DefUnit *current;
		struct D4S *next;
	} d4s;
	D4S *stack;
}; // end of DefUnitStack class

//@@@@@@@@@@@@@  Scene class @@@@@@@@@@@@@@@@@@@@
class Scene {
public:
	Light *light;  // attach linked list of lights here
	GeomObj * obj;  // attach the object list here
	Color3 background, ambient;
	int maxRecursionDepth;
	RGBpixmap pixmap[8]; //list of attached pixmaps
	float minReflectivity, minTransparency;
	bool isInShadow(Ray& f);
	Scene():light(NULL),obj(NULL),tail(NULL) { //default constructor
		currMtrl.setDefault();
		background.set(0,0,0.6f);
		ambient.set(0.1f,0.1f,0.1f);
		minReflectivity = 0.5;
		minTransparency = 0.5;
		maxRecursionDepth = 3;
	} 
	Scene(string fname) { //constructor that also reads in an SDL file
		Scene();
		read(fname);
	}
	void freeScene() { // release the object and light lists
		GeomObj *p = obj;
		while(p) {
			GeomObj* q = p;
			p = p->next;
			delete q;
		}
		Light * q = light;
		while(q) {
			Light* r = q;
			q = q->next;
			delete r;
		}
	}
	bool read(string fname);
	Color3 shade(Ray& ray);
	void getFirstHit(Ray& ray, Intersection& best);
	GeomObj* getObject();
	void drawSceneOpenGL() {
		for (GeomObj* p = obj; p; p=p->next)
			p->drawOpenGL();	// draw it
	}
	void makeLightsOpenGL() {
		int i=1;
		for (Light* p = light; p; p=p->next, i++) {
			float pos[4] = {p->pos.x, p->pos.y, p->pos.z, 0.0f};
			float color[4] = {p->pos.x, p->pos.y, p->pos.z};
			glLightfv(GL_LIGHT0+i, GL_POSITION, pos);
			//glLightfv(GL_LIGHT0+i, GL_AMBIENT, color);
			glLightfv(GL_LIGHT0+i, GL_DIFFUSE, color);
			//glLightfv(GL_LIGHT0+i, GL_SPECULAR, color);
			glEnable(GL_LIGHT0+i);
		}
	}
	void Scene::xfrmNormal(Vector3& res, Affine4& aff, Vector3 normal) {
		res.x = aff.m[0]*normal.x + aff.m[1]*normal.y + aff.m[2]*normal.z;
		res.y = aff.m[4]*normal.x + aff.m[5]*normal.y + aff.m[6]*normal.z;
		res.z = aff.m[8]*normal.x + aff.m[9]*normal.y + aff.m[10]*normal.z;
	}
private:
	// private stuff used only for reading a scene
	int line;
	int nextline;  
	ifstream  *file_in;
	strstream *f_in;
	strstream temp_fin;
	DefUnitStack *def_stack;
	GeomObj * tail; // tail of object list
	AffineStack affStk; // affine stack
	Material currMtrl;
	//<<<<<<<< methods >>>>>>>>>>>
	string nexttoken(void) { //########## nexttoken()
		char c;
		string token;
		int lastchar = 1;
		if (!f_in) {return(token); }
		if (f_in->eof()) {return(token);}
		while (f_in->get(c)) 
		{
			if (f_in->eof()) {
				return(token);
			}
			switch (c) {
			case '\n': nextline += 1;
			case ' ' :
			case '\t':
			case '\a':
			case '\b':
			case '\v':
			case '\f':
			case '\r': {
				if ( lastchar == 0 ) {return(token);}break; }
			case '{': {
				token = c; return(token); break;}
			case '}': {
				token = c;
				return(token);
				break; }
			case '!': {
				while ( c != '\n' && f_in->get(c)) {
				}
				nextline++; break;}
			default: {
				token = token + c;
				lastchar = 0;
				
				if ((f_in->peek() == '{') ||
					(f_in->peek() == '}') ) {
					if ( lastchar == 0 ) {
						return(token);
					} else {
						f_in->get(c);
						token = c;
						return(token);
					}
				}
				line = nextline;
						}
			}
		}
		return(" ");
	}
	float getFloat() { //############ getFloat()
		strstream tmp;
		float number;
		string str = nexttoken();  
		tmp << str;
		if(!(tmp >> number))
		{
			cerr << "Line " << line << ": error getting float" << endl;
			exit(-1);
		}
		else 
		{
			char t;
			if ( (tmp >> t ) ) 
			{
				cerr << "Line " << line << ": bum chars in number" << endl;
				exit(-1);    
			} 
		}
		return number;
	}
	bool isidentifier(string keyword) { //######## isidentifier
		string temp = keyword;
		if (!isalpha(temp[0])) return(false);
		for (unsigned int count = 1; count < temp.length(); count++) {
			if ((!isalnum(temp[count]))&& (temp[count]!='.')) return(false);
		}
		return(true);
	}
	void cleanUp() { //######### cleanUp -  release stuff after parsing file
		affStk.releaseAffines(); 		//delete affine stack
		def_stack->release();
		delete def_stack; // release the DefUnitStack memory
	}
	//+++++++++++++ TokenType +++++++++++++
	enum TokenType { IDENT, LIGHT, ROTATE, TRANSLATE, 
		SCALE, PUSH, POP, IDENTITYAFFINE, CIRCLE, 
		GLOBALAMBIENT, BACKGROUND, MINREFLECTIVITY,
		MINTRANSPARENCY, MAXRECURSIONDEPTH, CUBE, SPHERE, TORUS, PLANE,
		SQUARE, CYLINDER, CONE, TAPEREDCYLINDER,TETRAHEDRON, OCTAHEDRON, 
		DODECAHEDRON, ICOSAHEDRON, BUCKYBALL, TEAPOT, TRIANGLE, 
		DIAMOND, UNION, INTERSECTION, DIFFERENCEa, MAKEPIXMAP,
		MESH, DEFAULTMATERIALS, AMBIENT, DIFFUSE, SPECULAR,
		SPECULARFRACTION, SURFACEROUGHNESS, EMISSIVE, SPECULAREXPONENT,
		SPEEDOFLIGHT, TRANSPARENCY, REFLECTIVITY, PARAMETERS, TEXTURE,
		LFTCURLY, RGHTCURLY, DEF, USE, T_NULL, F_EOF, UNKNOWN };
	//<<<<<<<<<<<<<whichtoken >>>>>>>>>>>>>>
	TokenType whichtoken(string keyword) {
		string temp = keyword;
		if ( temp == "light" )            return (LIGHT);
		if ( temp == "rotate" )           return (ROTATE);
		if ( temp == "translate" )        return (TRANSLATE);
		if ( temp == "scale")             return (SCALE);
		if ( temp == "push")        	  return (PUSH);
		if ( temp == "pop")				  return (POP);
		if ( temp == "identityAffine")    return (IDENTITYAFFINE);
		if ( temp == "cube")              return (CUBE);
		if ( temp == "sphere")            return (SPHERE);
		if ( temp == "torus")             return (TORUS);
		if ( temp == "plane")             return (PLANE);
		if ( temp == "square")            return (SQUARE);
		if ( temp == "cylinder")          return (CYLINDER);
		if ( temp == "circle")			  return (CIRCLE);
		if ( temp == "taperedCylinder")   return (TAPEREDCYLINDER);
		if ( temp == "triangle")		  return (TRIANGLE);
		if ( temp == "cone")              return (CONE);
		if ( temp == "tetrahedron")       return (TETRAHEDRON);
		if ( temp == "octahedron")        return (OCTAHEDRON);
		if ( temp == "dodecahedron")      return (DODECAHEDRON);
		if ( temp == "icosahedron")       return (ICOSAHEDRON);
		if ( temp == "buckyball")         return (BUCKYBALL);
		if ( temp == "diamond")           return (DIAMOND);
		if ( temp == "teapot")			  return (TEAPOT);
		if ( temp == "union")             return (UNION);
		if ( temp == "intersection")      return (INTERSECTION);
		if ( temp == "difference")        return (DIFFERENCEa);
		if ( temp == "mesh")              return (MESH);
		if ( temp == "makePixmap")        return (MAKEPIXMAP);
		if ( temp == "defaultMaterials")  return (DEFAULTMATERIALS);
		if ( temp == "ambient")           return (AMBIENT);
		if ( temp == "diffuse")           return (DIFFUSE);
		if ( temp == "specular")          return (SPECULAR);
		if ( temp == "specularFraction")  return (SPECULARFRACTION);
		if ( temp == "surfaceRoughness")  return (SURFACEROUGHNESS);
		if ( temp == "emissive")          return (EMISSIVE);
		if ( temp == "specularExponent")  return (SPECULAREXPONENT);
		if ( temp == "speedOfLight")      return (SPEEDOFLIGHT);
		if ( temp == "transparency")      return (TRANSPARENCY);
		if ( temp == "reflectivity")      return (REFLECTIVITY);
		if ( temp == "parameters")        return (PARAMETERS);
		if ( temp == "texture")			  return (TEXTURE);
		if ( temp == "globalAmbient")	  return (GLOBALAMBIENT);
		if ( temp == "minReflectivity")	  return (MINREFLECTIVITY);
		if ( temp == "minTransparency")	  return (MINTRANSPARENCY);
		if ( temp == "maxRecursionDepth") return (MAXRECURSIONDEPTH);
		if ( temp == "background")        return (BACKGROUND);
		if ( temp == "{")                 return (LFTCURLY);
		if ( temp == "}")                 return (RGHTCURLY);
		if ( temp == "def")               return (DEF);
		if ( temp == "use")               return (USE);
		if ( temp == " " )                return (T_NULL);
		if ( isidentifier(temp) )         return (IDENT);
		cout << temp << ":" << (unsigned int)temp.length() << endl;
		return(UNKNOWN);
	} // end of whichtoken
}; // end of Scene.h
