#include "Shapes.h"
//<<<<<<<<<<<<<<<<<< hit() for Sphere >>>>>>>>>>>>>>>>>>>>>>>
bool Sphere::hit (Ray &r, Intersection& inter) { // intersect the ray with the generic sphere
	if (doScreenExtent && r.recurseLevel == 0)
		if (r.col < scrnExtnt.left || r.col > scrnExtnt.right || r.row < scrnExtnt.bott || r.row > scrnExtnt.top)
			return false; //misses screen extent	
	Ray genRay;
	xfrmRay(genRay,invTransf,r);			
	double A, B, C, discrim, disc_root, t1, t2;
	A = dot3D(genRay.dir, genRay.dir); 
	assert(A > 0); // fails if ray is sick 
	B = dot3D(genRay.start, genRay.dir);
	C = dot3D(genRay.start, genRay.start) - 1.0;
	discrim = B * B - A * C;
	if( discrim < 0.0) return false;   // ray missed: numHits = 0 
	disc_root = sqrt(discrim);		
	inter.numHits = 0;	// no hits yet	
	t1 = (-B - disc_root)/A;         // earlier hit 
	if (t1 > 0.00001) { // a pos hit time
		inter.hit[0].hitTime = t1;
		Point3 P(rayPos(genRay, t1));// position of hit spot
		inter.hit[0].hitPoint.set(P);
		inter.hit[0].hitObject = this;
		inter.hit[0].isEntering = 1;
		inter.hit[0].surface = 0;
		inter.hit[0].hitNormal.set(P); // vector is same as hit spot
		inter.numHits++; // got one hit so far
	}
	t2 = (-B + disc_root)/A; 		// second hit
	if (t2 > 0.00001) { // got a hit (first or second...)
		inter.hit[inter.numHits].hitTime = t2;
		Point3 P(rayPos(genRay, t2));  // position of hit spot
		inter.hit[inter.numHits].hitPoint.set(P);
		inter.hit[inter.numHits].isEntering = 0;
		inter.hit[inter.numHits].hitObject = this;
		inter.hit[inter.numHits].surface = 0;
		inter.hit[inter.numHits].hitNormal.set(P); //same as hit spot
		inter.numHits++;			// got another
	}
	return inter.numHits > 0;
}// end of hit() for Sphere

//<<<<<<<<<<<<<<<<<<<<<<<<<<<< readMesh >>>>>>>>>>>>>>>>>>>>>>>>
void Mesh::readMesh(string fname) {
	fstream inStream;
	inStream.open(fname.c_str(), ios ::in); //open needs a c-like string
	if(inStream.fail() || inStream.eof()) {
		cout << "can't open file or eof: " << fname << endl; 
		makeEmpty();
		return;
	}
	inStream >> numVerts >> numNorms >> numFaces;
	// make arrays for vertices, normals, and faces
	pt = new Point3[numVerts];        assert(pt != NULL);
	norm = new Vector3[numNorms];     assert(norm != NULL);
	face = new Face[numFaces];        assert(face != NULL);
	for(int i = 0; i < numVerts; i++) 	// read in the vertices
		inStream >> pt[i].x >> pt[i].y >> pt[i].z;
	for(int ii = 0; ii < numNorms; ii++) 	// read in the normals
		inStream >> norm[ii].x >> norm[ii].y >> norm[ii].z;
	for(int f = 0; f < numFaces; f++) { // read in face data
		inStream >> face[f].nVerts;
		int n = face[f].nVerts;
		face[f].vert = new VertexID[n]; assert(face[f].vert != NULL);
		for(int k = 0; k < n; k++) 		// read vertex indices for this face
			inStream >> face[f].vert[k].vertIndex;
		for(int kk = 0; kk < n; kk++) 		// read normal indices for this face
			inStream >> face[f].vert[kk].normIndex;
	}
	inStream.close();
} // end of readMesh

//<<<<<<<<<<<<<<<<<<<<<< drawMesh >>>>>>>>>>>>>>>>>>>>
void Mesh :: drawMesh() { // draw each face of this mesh using OpenGL: draw each polygon.
	if(isEmpty()) return; // mesh is empty
	for(int f = 0; f < numFaces; f++) {
		int n = face[f].nVerts;
		glBegin(GL_POLYGON);
		for(int v = 0; v < n; v++) {
			int in = face[f].vert[v].normIndex;
			assert(in >= 0 && in < numNorms);
			glNormal3f(norm[in].x, norm[in].y, norm[in].z);
			int iv = face[f].vert[v].vertIndex; assert(iv >= 0 && iv < numVerts);
			glVertex3f(pt[iv].x, pt[iv].y, pt[iv].z);
		}
		glEnd();	
	}
	glFlush();
}

//<<<<<<<<<<<<<<<<<<<<<<<<<<<< write mesh>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
void Mesh:: writeMesh(char * fname) { // write this mesh object into a new Chapter 6 format file.
	if(numVerts == 0 || numNorms   == 0 || numFaces == 0) return; //empty
	fstream outStream(fname, ios ::out); // open the output stream
	if(outStream.fail()) {cout << "can't make new file: " << fname << endl; return;}
	outStream << numVerts << " " << numNorms << " " << numFaces << "\n";
	// write the vertex and vertex normal list
	for(int i = 0; i < numVerts; i++)
		outStream << pt[i].x   << " " << pt[i].y   << " " << pt[i].z << "\n";
	for(int ii = 0; ii < numNorms; ii++)
		outStream  << norm[ii].x << " " << norm[ii].y << " " << norm[ii].z << "\n";
	// write the face data
	for(int f = 0; f < numFaces; f++) {
		int n = face[f].nVerts;
		outStream << n << "\n";
		for(int v = 0; v < n; v++)// write vertex indices for this face
			outStream << face[f].vert[v].vertIndex << " ";	outStream << "\n";
		for(int k = 0; k < n; k++)	// write normal indices for this face 
			outStream << face[f].vert[k].normIndex << " "; outStream << "\n";
	}
	outStream.close();
}
